import os


def get_data_dir():
    root = os.path.join(os.path.dirname(__file__), '../../')
    data_dir = os.path.join(root, 'data')
    return os.path.abspath(data_dir)


def get_data_file(filename):
    return os.path.abspath(os.path.join(get_data_dir(), filename))



from nlp_datasets.abstract_dataset import AbstractXYDataset, AbstractXYZDataset
from nlp_datasets.tokenizers import AbstractTokenizer, SpaceTokenizer
from nlp_datasets.tokenizers import NumbersTokenFilter, RegexTokenFilter
from nlp_datasets.tokenizers import TokenFilter, EmptyTokenFilter, LengthTokenFilter

from nlp_datasets.xy_dataset import XYSameFileDataset, XYSeparateFileDataset
from nlp_datasets.xyz_dataset import XYZSameFileDataset, XYZSeparateFileDataset

name = 'nlp_datasets'
__version__ = '1.2.1'



import abc
import logging
import os

import tensorflow as tf
from collections import Counter


class AbstractTokenizer(abc.ABC):
    """Tokenizer for language. The first line of vocab file is always `0   <UNK>`."""

    def __init__(self, config=None):
        default_config = self._get_default_config()
        if config:
            default_config.update(config)
        self.config = default_config

        self._special_tokens = [
            self.config['unk_token'],
            self.config['sos_token'],
            self.config['eos_token'],
            self.config['cls_token'],
            self.config['sep_token'],
            self.config['mask_token']
        ]

        self.counter = Counter()
        self._id2token_dict = None
        self._token2id_dict = None
        self._id2token_table = None
        self._token2id_table = None

    def _process_line(self, line, token_filters=None):
        """Process line by line in file.

        Args:
            line: A python str, line of file
            token_filters: An iterable of TokenFilter
        """
        raise NotImplementedError()

    @property
    def vocab_size(self):
        return len(self._token2id_dict)

    @property
    def unk_id(self):
        return 0

    @property
    def unk_token(self):
        return self.config.get('unk_token', '<UNK>')

    @property
    def sos_id(self):
        return self._token2id_dict.get(self.config['sos_token'], None)

    @property
    def sos_token(self):
        if not self.config['add_sos']:
            return None
        return self.config.get('sos_token', '<SOS>')

    @property
    def eos_id(self):
        return self._token2id_dict.get(self.config['eos_token'], None)

    @property
    def eos_token(self):
        if not self.config['add_eos']:
            return None
        return self.config.get('eos_token', '<EOS>')

    @property
    def cls_id(self):
        return self._token2id_dict.get(self.config['cls_token'], None)

    @property
    def cls_token(self):
        if not self.config['add_cls']:
            return None
        return self.config.get('cls_token', '[CLS]')

    @property
    def sep_id(self):
        return self._token2id_dict.get(self.config['sep_token'], None)

    @property
    def sep_token(self):
        if not self.config['add_sep']:
            return None
        return self.config.get('sep_token', '[SEP]')

    @property
    def mask_id(self):
        return self._token2id_dict.get(self.config['mask_token'], None)

    @property
    def mask_token(self):
        if not self.config['add_mask']:
            return None
        return self.config.get('mask_token', '[MASK]')

    @property
    def token2id_dict(self):
        return self._token2id_dict

    @property
    def id2token_dict(self):
        return self._id2token_dict

    def encode(self, tokens):
        """Encode string tokens to ids."""
        return self._token2id_table.lookup(tokens)

    def decode(self, ids):
        """Decode ids to string tokens."""
        return self._id2token_table.lookup(ids)

    def _reset(self):
        self.counter.clear()
        self._id2token_dict = dict()
        self._token2id_dict = dict()
        logging.info('Tokenizer reset.')

    def build_from_corpus(self, corpus_files, max_vocab_size=None, token_filters=None):
        """Build lookup table and vocab dict from corpus files.

        Args:
            corpus_files: An iterable of files
            max_vocab_size: Max size of vocab, not including special tokens like unk|sos|eos and so on.
            token_filters: An iterable of TokenFilter instances
        """
        self._reset()

        for f in corpus_files:
            if not os.path.exists(f):
                logging.warning('File %s does not exist.' % f)
                continue
            with open(f, mode='rt', encoding='utf8') as fin:
                for line in fin:
                    line = line.strip('\n')
                    if not line:
                        continue
                    self._process_line(line, token_filters)

        self._token2id_dict = {self.config['unk_token']: 0}

        # select most common vocabs
        for k, v in self.counter.most_common(max_vocab_size):
            self._token2id_dict[k] = len(self._token2id_dict)

        logging.info("Vocab size including unk is: %d" % len(self._token2id_dict))
        self._add_special_tokens(self._token2id_dict)
        logging.info("Vocab size including special tokens is: %d" % len(self._token2id_dict))

        for k, v in self._token2id_dict.items():
            self._id2token_dict[v] = k

        self._init_lookup_tables(self._token2id_dict, self._id2token_dict)
        logging.info("Build tokenizer from corpus files finished.")

    def build_from_vocab(self, vocab_file):
        """Build lookup table from vocab file."""
        self._reset()

        with open(vocab_file, mode='rt', encoding='utf8') as fin:
            for line in fin:
                line = line.strip('\n').strip()
                if not line:
                    continue
                if line in self._special_tokens:
                    continue
                self._token2id_dict[line] = len(self._token2id_dict)

        logging.info("Vocab size including unk is: %d" % len(self._token2id_dict))
        self._add_special_tokens(self._token2id_dict)
        logging.info("Vocab size including special tokens is: %d" % len(self._token2id_dict))

        for k, v in self._token2id_dict.items():
            self._id2token_dict[v] = k

        # init lookup tables
        self._init_lookup_tables(self._token2id_dict, self._id2token_dict)
        logging.info("Build tokenizer from vocab file finished.")

    def _add_special_tokens(self, token2id):
        if self.config['add_sos']:
            token2id[self.config['sos_token']] = len(token2id)
        if self.config['add_eos']:
            token2id[self.config['eos_token']] = len(token2id)
        if self.config['add_cls']:
            token2id[self.config['cls_token']] = len(token2id)
        if self.config['add_sep']:
            token2id[self.config['sep_token']] = len(token2id)
        if self.config['add_mask']:
            token2id[self.config['mask_token']] = len(token2id)

    def _init_lookup_tables(self, token2id_dict, id2token_dict):
        token2id_initializer = tf.lookup.KeyValueTensorInitializer(
            keys=list(token2id_dict.keys()),
            values=list(token2id_dict.values()),
            key_dtype=tf.dtypes.string,
            value_dtype=tf.dtypes.int64)
        self._token2id_table = tf.lookup.StaticHashTable(
            initializer=token2id_initializer,
            default_value=0,  # unk id
            name='token2id_lookup_table')

        id2token_initializer = tf.lookup.KeyValueTensorInitializer(
            keys=list(id2token_dict.keys()),
            values=list(id2token_dict.values()),
            key_dtype=tf.dtypes.int64,
            value_dtype=tf.dtypes.string)
        self._id2token_table = tf.lookup.StaticHashTable(
            initializer=id2token_initializer,
            default_value=self.config.get('unk_token', '<UNK>'),
            name='id2token_lookup_table')

    def save_to_vocab(self, output_file):
        with open(output_file, mode='wt', encoding='utf8') as fout:
            for k, v in self._id2token_dict.items():
                fout.write(str(v) + '\n')
        logging.info("Vocab saved in %s" % output_file)

    @staticmethod
    def _get_default_config():
        c = {
            'unk_token': '<UNK>',
            'sos_token': '<SOS>',
            'eos_token': '<EOS>',
            'cls_token': '[CLS]',
            'sep_token': '[SEP]',
            'mask_token': '[MASK]',
            'add_sos': True,
            'add_eos': True,
            'add_cls': False,  # whether append [CLS] token to the end of the vocab
            'add_sep': False,  # whether append [SEP] token to the end of the vocab
            'add_mask': False,  # whether append [MASK] token to the end of the vocab
        }
        return c



from .abstract_tokenizer import AbstractTokenizer
from .space_tokenizer import SpaceTokenizer
from .token_filter import TokenFilter, EmptyTokenFilter, LengthTokenFilter, NumbersTokenFilter, RegexTokenFilter



import unittest

import tensorflow as tf

from nlp_datasets.tokenizers import EmptyTokenFilter, SpaceTokenizer
from nlp_datasets.utils import data_dir_utils


class SpaceTokenizerTest(unittest.TestCase):

    def buildTokenizer(self):
        tokenizer = SpaceTokenizer()
        corpus = ['iwslt15.tst2013.100.en']
        corpus = [data_dir_utils.get_data_file(f) for f in corpus]
        tokenizer.build_from_corpus(corpus, token_filters=[EmptyTokenFilter()])
        return tokenizer

    def testTokenize(self):
        tokenizer = self.buildTokenizer()
        print(tokenizer.token2id_dict)
        print(tokenizer.id2token_dict)
        print(tokenizer.vocab_size)

    def testConvertTokens2Ids(self):
        tokenizer = self.buildTokenizer()
        print('token2 id dict: ', tokenizer.token2id_dict)
        words = tf.constant(['I', 'am', 'a', 'developer'])
        v = tokenizer.encode(words)
        print(v)

    def testConvertIds2Tokens(self):
        tokenizer = self.buildTokenizer()
        print('id2token dict: ', tokenizer.id2token_dict)
        ids = tf.constant([1, 0, 2, 3, 4], dtype=tf.dtypes.int64)
        v = tokenizer.decode(ids)
        print(v)

    def testSaveVocabFile(self):
        tokenizer = self.buildTokenizer()
        tokenizer.save_to_vocab(data_dir_utils.get_data_file('vocab.test.txt'))
        print(tokenizer.token2id_dict)
        print(tokenizer.id2token_dict)

    def testBuildFromVocab(self):
        print('============start build from vocab=============')
        tokenizer = SpaceTokenizer()
        tokenizer.build_from_vocab(data_dir_utils.get_data_file('vocab.test.txt'))
        print('token2id dict: ', tokenizer.token2id_dict)
        print('id2token dict: ', tokenizer.id2token_dict)
        words = tf.constant(['I', 'am', 'a', 'developer'])
        v0 = tokenizer.encode(words)
        print(v0)
        ids = tf.constant([1, 0, 2, 3, 4], dtype=tf.dtypes.int64)
        v1 = tokenizer.decode(ids)
        print(v1)
        print('============end build from vocab=============')


if __name__ == '__main__':
    unittest.main()



import tensorflow as tf

from nlp_datasets import AbstractXYDataset


class XYSameFileDataset(AbstractXYDataset):
    """Build dataset for examples that x and y are in the same file."""

    def _build_dataset(self, files):
        if isinstance(files, list):
            dataset = tf.data.Dataset.from_tensor_slices(files)
            dataset = dataset.flat_map(lambda f: tf.data.TextLineDataset(f).skip(self.config['skip_count']))
        else:
            dataset = tf.data.TextLineDataset(files)
        # repeat the dataset
        if self.config['repeat']:
            dataset = dataset.repeat(self.config['repeat'])
        return dataset

    def _build_dataset_from_files(self, files):
        """Build dataset from file(s).

        Args:
            files: A file or a list of file.

        Returns:
            A tf.data.Dataset object
        """
        dataset = self._build_dataset(files)
        # split x, y in one line
        dataset = dataset.map(
            lambda x: (tf.strings.split([x], sep=self.config['xy_sep']).values[0],
                       tf.strings.split([x], sep=self.config['xy_sep']).values[1]),
            num_parallel_calls=self.config['num_parallel_calls']
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def _build_dataset_from_files_for_predict(self, files):
        """Build dataset from file(s) for prediction.

        Args:
            files: A file or a list of file.

        Returns:
            A tf.data.Dataset object
        """
        dataset = self._build_dataset(files)
        dataset = dataset.map(
            lambda x: (tf.strings.split([x], sep=self.config['xy_sep']).values[0]),  # only takes x
            num_parallel_calls=self.config['num_parallel_calls']
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def _get_default_config(self):
        parent = super(XYSameFileDataset, self)._get_default_config()
        parent.update({
            'xy_sep': '@',  # split x and y
            'skip_count': 0,  # may useful when need to skip head line in files,
            'repeat': 1,  # repeating the dataset
        })
        return parent


class XYSeparateFileDataset(AbstractXYDataset):
    """Build dataset for examples that x and y are in different files."""

    def _build_dataset(self, files):
        if isinstance(files, list):
            dataset = tf.data.Dataset.from_tensor_slices(files)
            dataset = dataset.flat_map(lambda f: tf.data.TextLineDataset(f).skip(self.config['skip_count']))
        else:
            dataset = tf.data.TextLineDataset(files)
        # repeat the dataset
        if self.config['repeat']:
            dataset = dataset.repeat(self.config['repeat'])
        return dataset

    def _build_dataset_from_files(self, files):
        """Build dataset from file(s).

        Args:
            files: An iterable of tuple (x_file, y_file)

        Returns:
            A tf.data.Dataset object
        """
        x_files, y_files = files

        x_dataset = self._build_dataset(x_files)
        y_dataset = self._build_dataset(y_files)
        dataset = tf.data.Dataset.zip((x_dataset, y_dataset))
        return dataset

    def _build_dataset_from_files_for_predict(self, files):
        """Build dataset from file(s) for prediction.

        Args:
            files: An iterable of x_file

        Returns:
            A tf.data.Dataset object
        """
        x_files = files
        dataset = self._build_dataset(x_files)
        return dataset

    def _get_default_config(self):
        parent = super(XYSeparateFileDataset, self)._get_default_config()
        parent.update({
            'skip_count': 0,  # may useful when need to skip head line in files
            'repeat': 1,  # repeating the dataset
        })
        return parent



from nlp_datasets.tokenizers import AbstractTokenizer


class SpaceTokenizer(AbstractTokenizer):
    """Tokenize strings by SPACE."""

    def __init__(self, config=None):
        super(SpaceTokenizer, self).__init__(config)

    def _process_line(self, line, token_filters=None):
        for w in line.split(' '):
            w = w.strip()
            if not w:
                continue
            if token_filters:
                if any(f.drop(w) for f in token_filters):
                    continue
            self.counter[w] = self.counter.get(w, 0) + 1



import abc

import tensorflow as tf


class AbstractDataset(abc.ABC):
    """Dataset interface."""

    def __init__(self, x_tokenizer, y_tokenizer, config=None):
        """Constructor.

        Args:
            x_tokenizer: An instance of AbstractTokenizer for x language.
            y_tokenizer: An instance of AbstractTokenizer for y language.
            config: A python dict.
        """
        default_config = self._get_default_config()
        if config:
            default_config.update(config)
        self.config = default_config

        self.x_tokenizer = x_tokenizer
        self.y_tokenizer = y_tokenizer

    def build_train_dataset(self, train_files):
        """Build dataset for training."""
        raise NotImplementedError()

    def build_eval_dataset(self, eval_files):
        """Build dataset for evaluation."""
        raise NotImplementedError()

    def build_predict_dataset(self, predict_files):
        """Build dataset for prediction."""
        raise NotImplementedError()

    @staticmethod
    def _get_default_config():
        c = {
            'add_sos': True,
            'add_eos': True,
            'padding_by_eos': False,  # padding sequence by eos_token if True, unk_token else.
            'drop_remainder': False,
        }
        return c


class AbstractXYDataset(AbstractDataset):
    """Dataset interface for x-y sequence pair models."""

    def _shuffle_dataset(self, dataset):
        """Shuffle dataset."""
        dataset = dataset.shuffle(
            buffer_size=self.config['buffer_size'],  # buffer_size must be greater than total num of examples
            seed=self.config['seed'],
            reshuffle_each_iteration=self.config['reshuffle_each_iteration'])
        return dataset

    def _split_line(self, dataset):
        """Split sequence to list."""
        dataset = dataset.map(
            lambda x, y: (
                tf.strings.split([x], sep=self.config['sequence_sep']).values,
                tf.strings.split([y], sep=self.config['sequence_sep']).values),
            num_parallel_calls=self.config['num_parallel_calls']
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def _split_line_for_predict(self, dataset):
        dataset = dataset.map(
            lambda x: tf.strings.split([x], sep=self.config['sequence_sep']).values,
            num_parallel_calls=self.config['num_parallel_calls']
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def _filter_dataset(self, dataset):
        """Filter examples that are empty or too long."""
        dataset = dataset.filter(lambda x, y: tf.logical_and(tf.size(x) > 0, tf.size(y) > 0))
        x_max_len = self.config['x_max_len']
        if x_max_len > 0:
            dataset = dataset.filter(lambda x, y: tf.size(x) <= x_max_len)
        y_max_len = self.config['y_max_len']
        if y_max_len > 0:
            dataset = dataset.filter(lambda x, y: tf.size(y) <= y_max_len)
        return dataset

    def _filter_dataset_for_predict(self, dataset):
        dataset = dataset.filter(lambda x, y: tf.logical_and(tf.size(x) > 0, tf.size(y) > 0))
        x_max_len = self.config['x_max_len']
        if x_max_len > 0:
            dataset = dataset.filter(lambda x, y: tf.size(x) <= x_max_len)
        return dataset

    def _convert_tokens_to_ids(self, dataset):
        """Convert string tokens to ids."""
        dataset = dataset.map(
            lambda x, y: (self.x_tokenizer.encode(x), self.y_tokenizer.encode(y)),
            num_parallel_calls=self.config['num_parallel_calls']
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def _convert_tokens_to_ids_for_predict(self, dataset):
        dataset = dataset.map(
            lambda x: self.x_tokenizer.encode(x),
            num_parallel_calls=self.config['num_parallel_calls']
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def _add_special_tokens(self, dataset):
        """Attempt to add sos and eos tokens."""
        if self.config['add_sos']:
            x_sos_id = tf.constant(self.x_tokenizer.sos_id, dtype=tf.dtypes.int64)
            y_sos_id = tf.constant(self.y_tokenizer.sos_id, dtype=tf.dtypes.int64)
            dataset = dataset.map(
                lambda x, y: (tf.concat(([x_sos_id], x), axis=0), tf.concat(([y_sos_id], y), axis=0)),
                num_parallel_calls=self.config['num_parallel_calls']
            ).prefetch(self.config['prefetch_size'])
        if self.config['add_eos']:
            x_eos_id = tf.constant(self.x_tokenizer.eos_id, dtype=tf.dtypes.int64)
            y_eos_id = tf.constant(self.y_tokenizer.eos_id, dtype=tf.dtypes.int64)
            dataset = dataset.map(
                lambda x, y: (tf.concat((x, [x_eos_id]), axis=0), tf.concat((y, [y_eos_id]), axis=0)),
                num_parallel_calls=self.config['num_parallel_calls']
            ).prefetch(self.config['prefetch_size'])
        return dataset

    def _add_special_tokens_for_predict(self, dataset):
        if self.config['add_sos']:
            x_sos_id = tf.constant(self.x_tokenizer.sos_id, dtype=tf.dtypes.int64)
            dataset = dataset.map(
                lambda x,: tf.concat(([x_sos_id], x), axis=0),
                num_parallel_calls=self.config['num_parallel_calls']
            ).prefetch(self.config['prefetch_size'])
        if self.config['add_eos']:
            x_eos_id = tf.constant(self.x_tokenizer.eos_id, dtype=tf.dtypes.int64)
            dataset = dataset.map(
                lambda x: tf.concat((x, [x_eos_id]), axis=0),
                num_parallel_calls=self.config['num_parallel_calls']
            ).prefetch(self.config['prefetch_size'])
        return dataset

    def _padding_and_batching(self, dataset, batch_size, bucket_width=10):
        # batching
        def _batch_fn(x):
            x_max_len = self.config['x_max_len']
            x_padded_shape = x_max_len if x_max_len > 0 else None
            y_max_len = self.config['y_max_len']
            y_padded_shape = y_max_len if y_max_len > 0 else None
            if self.config['padding_by_eos']:
                x_padding_value = tf.constant(self.x_tokenizer.eos_id, dtype=tf.dtypes.int64)
                y_padding_value = tf.constant(self.y_tokenizer.eos_id, dtype=tf.dtypes.int64)
            else:
                x_padding_value = tf.constant(self.x_tokenizer.unk_id, dtype=tf.dtypes.int64)
                y_padding_value = tf.constant(self.y_tokenizer.unk_id, dtype=tf.dtypes.int64)

            x = x.padded_batch(
                batch_size=batch_size,
                padded_shapes=([x_padded_shape], [y_padded_shape]),
                padding_values=(x_padding_value, y_padding_value),
                drop_remainder=self.config['drop_remainder'])
            return x

        def _key_fn(x, y):
            # x -> src sequence, y -> target sequence, corresponding to dataset's two element
            bucket_id = tf.maximum(tf.size(x) // bucket_width, tf.size(y) // bucket_width)
            return tf.cast(bucket_id, tf.dtypes.int64)

        def _reduce_fn(k, x):
            # k -> unused key, x -> dataset
            return _batch_fn(x)

        bucket_fn = tf.data.experimental.group_by_window(
            key_func=_key_fn, reduce_func=_reduce_fn, window_size=batch_size)
        return dataset.apply(bucket_fn)

    def _padding_and_batching_for_predict(self, dataset, batch_size):
        x_max_len = self.config['x_max_len']
        x_padded_shape = x_max_len if x_max_len > 0 else None
        if self.config['padding_by_eos']:
            x_padding_value = tf.constant(self.x_tokenizer.eos_id, dtype=tf.dtypes.int64)
        else:
            x_padding_value = tf.constant(self.x_tokenizer.unk_id, dtype=tf.dtypes.int64)

        dataset = dataset.padded_batch(
            batch_size=batch_size,
            padded_shapes=[x_padded_shape],
            padding_values=x_padding_value)
        return dataset

    def build_train_dataset(self, train_files):
        """Build dataset for training.

        Args:
            train_files: An iterable of tuple (x_file, y_file)

        Returns:
            A tf.data.Dataset object
        """
        dataset = self._build_dataset_from_files(train_files)
        dataset = self._shuffle_dataset(dataset)
        dataset = self._split_line(dataset)
        dataset = self._filter_dataset(dataset)
        dataset = self._convert_tokens_to_ids(dataset)
        dataset = self._add_special_tokens(dataset)
        batch_size = self.config['train_batch_size']
        bucket_width = self.config['bucket_width']
        dataset = self._padding_and_batching(dataset, batch_size, bucket_width)
        return dataset

    def build_eval_dataset(self, eval_files):
        """Build dataset for evaluation.

        Args:
            eval_files: An iterable of tuple (x_file, y_file)

        Returns:
            A tf.data.Dataset object
        """
        dataset = self._build_dataset_from_files(eval_files)
        dataset = self._shuffle_dataset(dataset)
        dataset = self._split_line(dataset)
        dataset = self._filter_dataset(dataset)
        dataset = self._convert_tokens_to_ids(dataset)
        dataset = self._add_special_tokens(dataset)
        batch_size = self.config['eval_batch_size']
        bucket_size = self.config['bucket_width']
        dataset = self._padding_and_batching(dataset, batch_size, bucket_size)
        return dataset

    def build_predict_dataset(self, predict_files):
        """Build dataset for prediction.

        Args:
            predict_files: An iterable of x_file

        Returns:
            An tf.data.Dataset object
        """
        dataset = self._build_dataset_from_files_for_predict(predict_files)
        dataset = self._split_line_for_predict(dataset)
        dataset = self._convert_tokens_to_ids_for_predict(dataset)
        dataset = self._add_special_tokens_for_predict(dataset)
        dataset = self._padding_and_batching_for_predict(dataset, self.config['predict_batch_size'])
        return dataset

    def _build_dataset_from_files(self, files):
        """Build dataset from file(s) for training or evaluation mode.

        Args:
            files: An iterable of tuple (x_file, y_file)

        Returns:
            A tf.data.Dataset object
        """
        raise NotImplementedError()

    def _build_dataset_from_files_for_predict(self, files):
        """Build dataset from file(s) for prediction mode.

        Args:
            files: An iterable of x_file

        Returns:
            A tf.data.Dataset object
        """
        raise NotImplementedError()

    def _get_default_config(self):
        parent = super(AbstractXYDataset, self)._get_default_config()
        parent.update({
            'x_max_len': -1,
            'y_max_len': -1,
            'buffer_size': 10000000,
            'seed': None,
            'reshuffle_each_iteration': True,
            'sequence_sep': ' ',  # split x and y to list
            'prefetch_size': tf.data.experimental.AUTOTUNE,
            'num_parallel_calls': tf.data.experimental.AUTOTUNE,
            'train_batch_size': 32,
            'eval_batch_size': 32,
            'predict_batch_size': 32,
            'bucket_width': 10,
        })
        return parent


class AbstractXYZDataset(AbstractDataset):

    def _shuffle_dataset(self, dataset):
        """Shuffle dataset."""
        dataset = dataset.shuffle(
            buffer_size=self.config['buffer_size'],  # buffer_size must be greater than total num of examples
            seed=self.config['seed'],
            reshuffle_each_iteration=self.config['reshuffle_each_iteration'])
        return dataset

    def _split_line(self, dataset):
        dataset = dataset.map(
            lambda x, y, z: (tf.strings.split([x], sep=self.config['sequence_sep']).values,
                             tf.strings.split([y], sep=self.config['sequence_sep']).values,
                             self._normalize_z(z)),
            num_parallel_calls=self.config['num_parallel_calls']
        ).prefetch(self.config['prefetch_size'])
        return dataset

    @tf.function
    def _normalize_z(self, z):
        raise NotImplementedError()

    def _split_line_for_predict(self, dataset):
        dataset = dataset.map(
            lambda x, y: (tf.strings.split([x], sep=self.config['sequence_sep']).values,
                          tf.strings.split([y], sep=self.config['sequence_sep']).values,),
            num_parallel_calls=self.config['num_parallel_calls']
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def _filter_dataset(self, dataset):
        dataset = dataset.filter(lambda x, y, z: tf.logical_and(tf.size(x) > 0, tf.size(y) > 0))
        x_max_len = self.config['x_max_len']
        if x_max_len > 0:
            dataset = dataset.filter(lambda x, y, z: (tf.size(x) <= x_max_len))
        y_max_len = self.config['y_max_len']
        if y_max_len > 0:
            dataset = dataset.filter(lambda x, y, z: (tf.size(y) <= y_max_len))
        return dataset

    def _convert_tokens_to_ids(self, dataset):
        dataset = dataset.map(
            lambda x, y, z: (self.x_tokenizer.encode(x), self.y_tokenizer.encode(y), z),
            num_parallel_calls=self.config['num_parallel_calls'],
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def _convert_tokens_to_ids_for_predict(self, dataset):
        dataset = dataset.map(
            lambda x, y: (self.x_tokenizer.encode(x), self.y_tokenizer.encode(y)),
            num_parallel_calls=self.config['num_parallel_calls'],
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def _add_special_tokens(self, dataset):
        if self.config['add_sos']:
            x_sos = tf.constant(self.x_tokenizer.sos_id, dtype=tf.dtypes.int64)
            y_sos = tf.constant(self.y_tokenizer.sos_id, dtype=tf.dtypes.int64)
            dataset = dataset.map(
                lambda x, y, z: (tf.concat(([x_sos], x), axis=0),
                                 tf.concat(([y_sos], y), axis=0),
                                 z),
                num_parallel_calls=self.config['num_parallel_calls']
            ).prefetch(self.config['prefetch_size'])
        if self.config['add_eos']:
            x_eos = tf.constant(self.x_tokenizer.eos_id, dtype=tf.dtypes.int64)
            y_eos = tf.constant(self.y_tokenizer.eos_id, dtype=tf.dtypes.int64)
            dataset = dataset.map(
                lambda x, y, z: (tf.concat((x, [x_eos]), axis=0),
                                 tf.concat((y, [y_eos]), axis=0),
                                 z),
                num_parallel_calls=self.config['num_parallel_calls']
            ).prefetch(self.config['prefetch_size'])
        return dataset

    def _add_special_tokens_for_predict(self, dataset):
        if self.config['add_sos']:
            x_sos = tf.constant(self.x_tokenizer.sos_id, dtype=tf.dtypes.int64)
            y_sos = tf.constant(self.y_tokenizer.sos_id, dtype=tf.dtypes.int64)
            dataset = dataset.map(
                lambda x, y: (tf.concat(([x_sos], x), axis=0),
                              tf.concat(([y_sos], y), axis=0),),
                num_parallel_calls=self.config['num_parallel_calls']
            ).prefetch(self.config['prefetch_size'])
        if self.config['add_eos']:
            x_eos = tf.constant(self.x_tokenizer.eos_id, dtype=tf.dtypes.int64)
            y_eos = tf.constant(self.y_tokenizer.eos_id, dtype=tf.dtypes.int64)
            dataset = dataset.map(
                lambda x, y: (tf.concat((x, [x_eos]), axis=0),
                              tf.concat((y, [y_eos]), axis=0),),
                num_parallel_calls=self.config['num_parallel_calls']
            ).prefetch(self.config['prefetch_size'])
        return dataset

    def _padding_and_batching(self, dataset, batch_size, bucket_width=10):
        # batching
        def _batch_fn(x):
            x_max_len = self.config['x_max_len']
            x_padded_shape = x_max_len if x_max_len > 0 else None
            y_max_len = self.config['y_max_len']
            y_padded_shape = y_max_len if y_max_len > 0 else None
            # padding sequence by sos or unk
            if self.config['padding_by_eos']:
                x_padding_value = tf.constant(self.x_tokenizer.eos_id, dtype=tf.dtypes.int64)
                y_padding_value = tf.constant(self.y_tokenizer.eos_id, dtype=tf.dtypes.int64)
            else:
                x_padding_value = tf.constant(self.x_tokenizer.unk_id, dtype=tf.dtypes.int64)
                y_padding_value = tf.constant(self.y_tokenizer.unk_id, dtype=tf.dtypes.int64)

            x = x.padded_batch(
                batch_size=batch_size,
                padding_values=(x_padding_value, y_padding_value, tf.constant(0, dtype=tf.dtypes.int64)),
                padded_shapes=([x_padded_shape], [y_padded_shape], []),
                drop_remainder=self.config['drop_remainder']
            ).prefetch(self.config['prefetch_size'])

            # map (x,y,z) to ((x,y),z) cause (x,y) are two inputs of model.
            x = x.map(
                lambda x, y, z: ((x, y), z),
                num_parallel_calls=self.config['num_parallel_calls']
            ).prefetch(self.config['prefetch_size'])
            return x

        def _key_fn(x, y, z):
            # x -> src sequence, y -> tgt sequence, z -> label
            bucket_id = tf.maximum(tf.size(x) // bucket_width, tf.size(y) // bucket_width)
            return tf.cast(bucket_id, tf.dtypes.int64)

        def _reduce_fn(k, x):
            # k -> unused key, x -> dataset
            return _batch_fn(x)

        bucket_fn = tf.data.experimental.group_by_window(
            key_func=_key_fn, reduce_func=_reduce_fn, window_size=batch_size)
        return dataset.apply(bucket_fn)

    def _padding_and_batching_for_predict(self, dataset, batch_size):
        x_max_len = self.config['x_max_len']
        x_padded_shape = x_max_len if x_max_len > 0 else None
        y_max_len = self.config['y_max_len']
        y_padded_shape = y_max_len if y_max_len > 0 else None
        if self.config['padding_by_eos']:
            x_padding_value = tf.constant(self.x_tokenizer.eos_id, dtype=tf.dtypes.int64)
            y_padding_value = tf.constant(self.y_tokenizer.eos_id, dtype=tf.dtypes.int64)
        else:
            x_padding_value = tf.constant(self.x_tokenizer.unk_id, dtype=tf.dtypes.int64)
            y_padding_value = tf.constant(self.y_tokenizer.unk_id, dtype=tf.dtypes.int64)

        dataset = dataset.padded_batch(
            batch_size=batch_size,
            padding_values=(x_padding_value, y_padding_value),
            padded_shapes=([x_padded_shape], [y_padded_shape])
        ).prefetch(self.config['prefetch_size'])

        # map (x,y) to ((x,y)) cause (x,y) are two inputs of model.
        dataset = dataset.map(
            lambda x, y: ((x, y),),
            num_parallel_calls=self.config['num_parallel_calls']
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def build_train_dataset(self, train_files):
        dataset = self._build_dataset_from_files(train_files)
        dataset = self._shuffle_dataset(dataset)
        dataset = self._split_line(dataset)
        dataset = self._filter_dataset(dataset)
        dataset = self._convert_tokens_to_ids(dataset)
        dataset = self._add_special_tokens(dataset)
        batch_size = self.config['train_batch_size']
        bucket_width = self.config['bucket_width']
        dataset = self._padding_and_batching(dataset, batch_size, bucket_width)
        return dataset

    def build_eval_dataset(self, eval_files):
        dataset = self._build_dataset_from_files(eval_files)
        dataset = self._shuffle_dataset(dataset)
        dataset = self._split_line(dataset)
        dataset = self._filter_dataset(dataset)
        dataset = self._convert_tokens_to_ids(dataset)
        dataset = self._add_special_tokens(dataset)
        batch_size = self.config['eval_batch_size']
        bucket_width = self.config['bucket_width']
        dataset = self._padding_and_batching(dataset, batch_size, bucket_width)
        return dataset

    def build_predict_dataset(self, predict_files):
        dataset = self._build_dataset_from_files_for_predict(predict_files)
        dataset = self._split_line_for_predict(dataset)
        dataset = self._convert_tokens_to_ids_for_predict(dataset)
        dataset = self._add_special_tokens_for_predict(dataset)
        dataset = self._padding_and_batching_for_predict(dataset, self.config['predict_batch_size'])
        return dataset

    def _build_dataset_from_files(self, files):
        raise NotImplementedError()

    def _build_dataset_from_files_for_predict(self, files):
        raise NotImplementedError()

    def _get_default_config(self):
        parent = super(AbstractXYZDataset, self)._get_default_config()
        parent.update({
            'x_max_len': -1,
            'y_max_len': -1,
            'buffer_size': 10000000,
            'seed': None,
            'reshuffle_each_iteration': True,
            'sequence_sep': ' ',
            'prefetch_size': tf.data.experimental.AUTOTUNE,
            'num_parallel_calls': tf.data.experimental.AUTOTUNE,
            'train_batch_size': 32,
            'eval_batch_size': 32,
            'predict_batch_size': 32,
            'bucket_width': 10,
        })
        return parent



import tensorflow as tf

from nlp_datasets import AbstractXYZDataset


class XYZSameFileDataset(AbstractXYZDataset):

    def _build_dataset(self, files):
        if isinstance(files, list):
            dataset = tf.data.Dataset.from_tensor_slices(files)
            dataset = dataset.flat_map(lambda f: tf.data.TextLineDataset(f).skip(self.config['skip_count']))
        else:
            dataset = tf.data.TextLineDataset(files)
        return dataset

    @tf.function
    def _normalize_z(self, z):
        if tf.equal(z, '0 1'):
            return tf.constant(1, dtype=tf.dtypes.int64)
        if tf.equal(z, '1 0'):
            return tf.constant(0, dtype=tf.dtypes.int64)
        if tf.equal(z, '0'):
            return tf.constant(0, dtype=tf.dtypes.int64)
        if tf.equal(z, '1'):
            return tf.constant(1, dtype=tf.dtypes.int64)
        return tf.cast(tf.strings.to_number(z), dtype=tf.dtypes.int64)

    def _build_dataset_from_files(self, files):
        dataset = self._build_dataset(files)
        dataset = dataset.filter(lambda x: tf.size(tf.strings.split([x], sep=self.config['xyz_sep']).values) == 3)
        # split x, y, z in one line
        dataset = dataset.map(
            lambda x: (tf.strings.split([x], sep=self.config['xyz_sep']).values[0],
                       tf.strings.split([x], sep=self.config['xyz_sep']).values[1],
                       tf.strings.split([x], sep=self.config['xyz_sep']).values[2]),
            num_parallel_calls=self.config['num_parallel_calls']
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def _build_dataset_from_files_for_predict(self, predict_files):
        dataset = self._build_dataset(predict_files)
        dataset = dataset.filter(lambda x: tf.size(tf.strings.split([x], sep=self.config['xyz_sep'])) == 3)
        # split x, y, z in one line, and take x, y only!
        dataset = dataset.map(
            lambda x: (tf.strings.split([x], sep=self.config['xyz_sep']).values[0],
                       tf.strings.split([x], sep=self.config['xyz_sep']).values[1]),
            num_parallel_calls=self.config['num_parallel_calls']
        ).prefetch(self.config['prefetch_size'])
        return dataset

    def _get_default_config(self):
        parent = super(XYZSameFileDataset, self)._get_default_config()
        parent.update({
            'xyz_sep': '@',
            'skip_count': 0,
        })
        return parent


class XYZSeparateFileDataset(AbstractXYZDataset):

    def _build_dataset(self, files):
        if isinstance(files, list):
            dataset = tf.data.Dataset.from_tensor_slices(files)
            dataset = dataset.flat_map(lambda f: tf.data.TextLineDataset(f).skip(self.config['skip_count']))
        else:
            dataset = tf.data.TextLineDataset(files)
        return dataset

    @tf.function
    def _normalize_z(self, z):
        if tf.equal(z, b'0 1'):
            return tf.constant(1, dtype=tf.dtypes.int64)
        if tf.equal(z, b'1 0'):
            return tf.constant(0, dtype=tf.dtypes.int64)
        if tf.equal(z, b'0'):
            return tf.constant(0, dtype=tf.dtypes.int64)
        if tf.equal(z, b'1'):
            return tf.constant(1, dtype=tf.dtypes.int64)
        return tf.cast(tf.strings.to_number(z), dtype=tf.dtypes.int64)

    def _build_dataset_from_files(self, train_files):
        x_files, y_files, z_files = train_files
        x_dataset = self._build_dataset(x_files)
        y_dataset = self._build_dataset(y_files)
        z_dataset = self._build_dataset(z_files)
        dataset = tf.data.Dataset.zip((x_dataset, y_dataset, z_dataset))
        return dataset

    def _build_dataset_from_files_for_predict(self, predict_files):
        x_files, y_files = predict_files
        x_dataset = self._build_dataset(x_files)
        y_dataset = self._build_dataset(y_files)
        dataset = tf.data.Dataset.zip((x_dataset, y_dataset))
        return dataset

    def _get_default_config(self):
        parent = super(XYZSeparateFileDataset, self)._get_default_config()
        parent.update({
            'xyz_sep': '@',
            'skip_count': 0,
        })
        return parent



import abc
import re

NUMBERS_PATTERN = re.compile('^[0-9]+$')


class TokenFilter(abc.ABC):

    def drop(self, token):
        raise NotImplementedError()


class EmptyTokenFilter(TokenFilter):

    def drop(self, token):
        if not token or not token.strip():
            return True
        return False


class NumbersTokenFilter(TokenFilter):

    def drop(self, token):
        m = NUMBERS_PATTERN.match(token)
        if m:
            return True
        return False


class LengthTokenFilter(TokenFilter):

    def __init__(self, max_len):
        self.max_len = max(0, max_len)

    def drop(self, token):
        return len(token) > self.max_len


class RegexTokenFilter(TokenFilter):

    def __init__(self, regex):
        self.pattern = re.compile(regex)

    def drop(self, token):
        m = self.pattern.match(token)
        if m:
            return True
        return False



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nlp_datasets",
    version="1.2.1",
    description="A dataset utils repository based on tf.data. For tensorflow 2.x only!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/luozhouyang/nlp-datasets",
    author="ZhouYang Luo",
    author_email="zhouyang.luo@gmail.com",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[

    ],
    license="MIT License",
    classifiers=(
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    )
)



from nlp_datasets import XYSameFileDataset, XYSeparateFileDataset
import unittest
from nlp_datasets.utils import data_dir_utils
from nlp_datasets.tokenizers import SpaceTokenizer


class XYDatasetTest(unittest.TestCase):

    @staticmethod
    def buildXTokenizer():
        t = SpaceTokenizer()
        t.build_from_corpus([data_dir_utils.get_data_file('iwslt15.tst2013.100.en')])
        return t

    @staticmethod
    def buildYTokenizer():
        t = SpaceTokenizer()
        t.build_from_corpus([data_dir_utils.get_data_file('iwslt15.tst2013.100.vi')])
        return t

    def testXYSameFileDataset(self):
        x_tokenizer = self.buildXTokenizer()
        y_tokenizer = self.buildYTokenizer()
        config = {
            'xy_sep': '@',
            'sequence_sep': ' ',
            'x_max_len': -1,
            'y_max_len': -1,
            'repeat': 10,
        }
        ds = XYSameFileDataset(x_tokenizer, y_tokenizer, config=config)
        files = [data_dir_utils.get_data_file('iwslt15.tst2013.100.envi')]
        train_dataset = ds.build_train_dataset(train_files=files)
        v = next(iter(train_dataset))
        print(v)
        print('=' * 50)
        eval_dataset = ds.build_eval_dataset(files)
        v = next(iter(eval_dataset))
        print(v)
        print('=' * 50)
        predict_dataset = ds.build_predict_dataset(files)
        v = next(iter(predict_dataset))
        print(v)

    def testXYSeparateFileDataset(self):
        x_tokenizer = self.buildXTokenizer()
        y_tokenizer = self.buildYTokenizer()
        config = {
            'sequence_sep': ' ',
            'x_max_len': -1,
            'y_max_len': -1,
            'repeat': 10,
        }
        ds = XYSeparateFileDataset(x_tokenizer, y_tokenizer, config)
        files = ([data_dir_utils.get_data_file('iwslt15.tst2013.100.en')],
                 [data_dir_utils.get_data_file('iwslt15.tst2013.100.vi')])
        train_dataset = ds.build_train_dataset(train_files=files)
        v = next(iter(train_dataset))
        print(v)
        print('=' * 50)
        eval_dataset = ds.build_eval_dataset(files)
        v = next(iter(eval_dataset))
        print(v)
        print('=' * 50)
        predict_dataset = ds.build_predict_dataset([data_dir_utils.get_data_file('iwslt15.tst2013.100.en')])
        v = next(iter(predict_dataset))
        print(v)


if __name__ == '__main__':
    unittest.main()



from nlp_datasets import XYZSameFileDataset, XYZSeparateFileDataset

from nlp_datasets.utils import data_dir_utils
import unittest
from nlp_datasets.tokenizers import SpaceTokenizer


class XYZDatasetTest(unittest.TestCase):

    @staticmethod
    def buildXTokenizer():
        t = SpaceTokenizer()
        t.build_from_corpus([data_dir_utils.get_data_file('dssm.query.txt')])
        return t

    @staticmethod
    def buildYTokenizer():
        t = SpaceTokenizer()
        t.build_from_corpus([data_dir_utils.get_data_file('dssm.doc.txt')])
        return t

    def testXYZSameFileDataset(self):
        x_tokenizer = self.buildXTokenizer()
        y_tokenizer = self.buildYTokenizer()
        config = {
            'xyz_sep': '@',
            'sequence_sep': ' ',
            'x_max_len': -1,
            'y_max_len': -1,
            'train_batch_size': 2,
            'eval_batch_size': 2,
            'predict_batch_size': 1
        }
        ds = XYZSameFileDataset(x_tokenizer, y_tokenizer, config=config)
        files = [data_dir_utils.get_data_file('dssm.query.doc.label.txt')]
        train_dataset = ds.build_train_dataset(files)
        v = next(iter(train_dataset))
        print(v)
        print('=' * 50)
        eval_dataset = ds.build_eval_dataset(files)
        v = next(iter(eval_dataset))
        print(v)
        print('=' * 50)
        predict_dataset = ds.build_predict_dataset(files)
        v = next(iter(predict_dataset))
        print(v)

    def testXYZSeparateFileDataset(self):
        x_tokenizer = self.buildXTokenizer()
        y_tokenizer = self.buildYTokenizer()
        config = {
            'xyz_sep': '@',
            'sequence_sep': ' ',
            'x_max_len': -1,
            'y_max_len': -1,
            'train_batch_size': 2,
            'eval_batch_size': 2,
            'predict_batch_size': 1
        }
        ds = XYZSeparateFileDataset(x_tokenizer, y_tokenizer, config=config)
        files = ([data_dir_utils.get_data_file('dssm.query.txt')],
                 [data_dir_utils.get_data_file('dssm.doc.txt')],
                 [data_dir_utils.get_data_file('dssm.label.txt')])
        train_dataset = ds.build_train_dataset(files)
        v = next(iter(train_dataset))
        print(v)
        print('=' * 50)
        eval_dataset = ds.build_eval_dataset(files)
        v = next(iter(eval_dataset))
        print(v)
        print('=' * 50)
        predict_dataset = ds.build_predict_dataset(
            ([data_dir_utils.get_data_file('dssm.query.txt')], [data_dir_utils.get_data_file('dssm.doc.txt')]))
        v = next(iter(predict_dataset))
        print(v)


if __name__ == '__main__':
    unittest.main()
